/* -*- C++ -*- *************************************************************
 *
 *  lsf.cpp, Mon Aug 23 10:09:08 MSD 2004
 *  Copyright (C) 2004 Soloviev Alexei <solovjev@cv.jinr.ru>, Alexei Stadnik
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lsf.h"

#include <algorithm>
#include <cctype>
#include <cmath>
#include <iomanip>
using namespace std;

ChiSquare::ChiSquare()
  : TMinuit ( 0 ),
    __m_ptr ( 0 ), 
    __x_ptr ( 0 ),  __y_ptr ( 0 ), __dy_ptr ( 0 ), 
    __n ( 0 ),
    __fun_ptr ( &ChiSquare::standard ), 
    __chi2 ( 0 ), 
    __weight(), __norm ( 0 ), C_TUKEY ( 5 ), 
    __strategy ( *__strategies ) 
{
  /** MINUIT initialization */
  SetFCN ( fcn );
}

ChiSquare::~ChiSquare() {}

/** Model access (read) */
const Model & ChiSquare::model() const
{
  return *__m_ptr;
}

/** Model access (write) */
void ChiSquare::model ( Model * m )
{
  __m_ptr = m;
  BuildArrays ( model().size() );
}

/** Data access (write) */
void ChiSquare::data ( const double * x, const double * y, const double * dy, 
		       const unsigned int n )
{
  __x_ptr = x;
  __y_ptr = y;
  __dy_ptr = dy;
  __n = n;
}

/** Chi-square (to be minimized) */
const double & ChiSquare::value() const
{
  return (this->*__fun_ptr)();
}

/** Chi-square norm */
const double ChiSquare::norm() const
{
  if ( __norm ) 
    return __norm - model().size();
  else 
    return __n - model().size();
}

/** Standard Chi-square value */
const double & ChiSquare::standard() const
{
  __chi2 = 0;
  register double di;
  const double *x_ptr = __x_ptr, *y_ptr = __y_ptr, *dy_ptr = __dy_ptr;
  for ( register unsigned int i = 0; i < __n; ++i ) {
    if ( *dy_ptr ) {
      di = ( (*__m_ptr)(*x_ptr++) - *y_ptr++ ) / *dy_ptr++;
      __chi2 += ( di *= di );
    }
  }
  return __chi2;
}

/** Weighted Chi-square value */
const double & ChiSquare::weighted() const
{
  __chi2 = 0;
  if ( __norm ) {
    register double di;
    const double *x_ptr = __x_ptr, *y_ptr = __y_ptr, *dy_ptr = __dy_ptr;
    for ( register unsigned int i = 0; i < __n; ++i ) {
      if ( *dy_ptr && __weight[i] ) {
	di = ( (*__m_ptr)(*x_ptr++) - *y_ptr++ ) / *dy_ptr++;
	__chi2 += ( ( di *= di ) *= __weight[i] );
      }
    }
  }
  return __chi2;
}

/** Calculate Tukey's bi-square weights for weighted LSF value */
void ChiSquare::set_weights()
{
  register double di;
  double range = value()/norm() * C_TUKEY;
  __norm = 0;
  const double *x_ptr = __x_ptr, *y_ptr = __y_ptr, *dy_ptr = __dy_ptr;
  for ( register unsigned int i = 0; i < __n; ++i ) {
    di = ( (*__m_ptr)(*x_ptr++) - *y_ptr++ ) / *dy_ptr++;
    if ( abs ( di ) < range ) {
      __weight[i] = ( ( ( ( di /= range ) *= di ) -= 1 ) *= di );
      __norm += __weight[i];
    } else {
      __weight[i] = 0;
    }
  }
}

/** Set weights = 1 for weighted LSF value 
    (should be called before any weights usage!) */
void ChiSquare::init_weights()
{
  if ( __weight.size() != __n ) {
    __weight.clear();
    __weight.resize ( __n, 1 );
  } else {
    fill ( __weight.begin(), __weight.end(), 1 );
  }
  __norm = __n;
}

/** Current MINUIT minimizer */
const string & ChiSquare::strategy() const
{
  return __strategy;
}

/** Setup MINUIT minimizer. Strategies are: 
    MINIMIZE (default), MIGRAD, SCAN, SEEK, SIMPLEX */
bool ChiSquare::strategy ( const string & method )
{
  string __method;
  transform ( method.begin(), method.end(), 
	      back_inserter ( __method ), ::toupper );
  if ( find ( __strategies, __strategies+5, __method ) != __strategies+5 ) {
    __strategy = __method;
    return true;
  }
  return false;
}

/** MINUIT interface */
void ChiSquare::minimize ( bool robust, const int verbose )
{
  static int ierflg;
  __lsf_ptr = this;

  /** What should be minimized? */
  if ( robust ) {
    __fun_ptr = &ChiSquare::weighted;
    init_weights();
  } else {
    __fun_ptr = &ChiSquare::standard;
  }

  /** Here we go! */
  set_mn_printout ( verbose );
  set_mn_parms();
  mncomd ( __strategy.c_str(), ierflg );
  if ( robust ) {
    double __chi2prev = __chi2;
    for ( C_TUKEY = 5; C_TUKEY > 1; C_TUKEY *= 0.9 ) {
      set_weights();
      mncomd ( __strategy.c_str(), ierflg );
      if ( abs ( __chi2 - __chi2prev ) / __chi2 < 0.01 ) break;
      __chi2prev = __chi2;
    }
  }
  get_mn_parms();
}

/** ChiSquare output */
ostream & operator<< ( ostream & out, const ChiSquare & chi )
{
  out.setf(ios::fixed,ios::floatfield);
  out.precision(4);

  int line_title = 16;
  int line_value = 24;
  int line_error = 12;
  int line_length = line_title + line_value + line_error;

  out << setw(line_length) << setfill('_') << "" << endl;

  out.fill('.');
  out.setf(ios::left,ios::adjustfield);
  out << setw(line_title) <<  "Model";
  out.setf(ios::right,ios::adjustfield);
  out << setw(line_value+line_error) << chi.model().name().c_str() << endl;

  out << setw(line_length) << setfill('_') << "" << endl;

  out.fill('.');
  for ( int i = 0; i < chi.model().size(); ++i) {
    out.setf(ios::left,ios::adjustfield);
    out << setw(line_title) << chi.model().name(i).c_str();
    out.setf(ios::right,ios::adjustfield);
    out << setw(line_value) << chi.model()[i];
    out << " +- ";
    out << chi.model().error(i) << endl;
  }

  out << setw(line_length) << setfill('_') << "" << endl;

  out.fill('.');
  out.setf(ios::left,ios::adjustfield);
  out << setw(line_title) << "Chi square";
  out.setf(ios::right,ios::adjustfield);
  out << setw(line_value+line_error) << chi.value()/chi.norm() << endl;

  out << setw(line_length) << setfill('_') << "" << endl;

  return out;
}

QTextStream & operator<< ( QTextStream & out, const ChiSquare & chi )
{
    out.setRealNumberNotation(QTextStream::FixedNotation);
    out.setRealNumberPrecision(4);

    int line_title = 16;
    int line_value = 24;
    int line_error = 12;
    int line_length = line_title + line_value + line_error + 1;

    out.setFieldWidth(line_length);
    out.setPadChar('_');
    out << endl;

    out.setPadChar('.');
    out.setFieldAlignment(QTextStream::AlignLeft);
    out.setFieldWidth(line_title);
    out << "Model";
    out.setFieldAlignment(QTextStream::AlignRight);
    out.setFieldWidth(line_value+line_error);
    out << chi.model().name().c_str();
    out.setPadChar(' ');
    out << endl;

    out.setFieldWidth(line_length);
    out.setPadChar('_');
    out << endl;

    for ( int i = 0; i < chi.model().size(); ++i) {
        out.setPadChar('.');
        out.setFieldAlignment(QTextStream::AlignLeft);
        out.setFieldWidth(line_title);
        out << chi.model().name(i).c_str();
        out.setFieldAlignment(QTextStream::AlignRight);
        out.setFieldWidth(line_value);
        out << chi.model()[i];
        out.setFieldWidth(0);
        out << " +- ";
        out << chi.model().error(i);
        out.setPadChar(' ');
        out << endl;
    }

    out.setFieldWidth(line_length);
    out.setPadChar('_');
    out << endl;

    out.setPadChar('.');
    out.setFieldAlignment(QTextStream::AlignLeft);
    out.setFieldWidth(line_title);
    out << "Chi square";
    out.setFieldAlignment(QTextStream::AlignRight);
    out.setFieldWidth(line_value+line_error) ;
    out << chi.value()/chi.norm();
    out.setPadChar(' ');
    out << endl;

    out.setFieldWidth(line_length);
    out.setPadChar('_');
    out << endl;

    return out;
}
