/* -*- C++ -*- *************************************************************
 *
 *  lsf.h, Mon Aug 23 10:08:25 MSD 2004
 *  Copyright (C) 2004 Soloviev Alexei <solovjev@cv.jinr.ru>, Alexei Stadnik
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LSF_H
#define LSF_H

#include "TMinuit.h"
#include "model.h"

#include <string>
#include <iostream>
using namespace std;

/***************************************************************************
 *                                                                         *
 *  Chi-square is a function to be minimized while Least-Squares Fit (LSF) *
 *                                                                         *
 *  Usage:                                                                 *
 *                                                                         *
 *  // create object:                                                      *
 *  ChiSquare chi_square;                                                  *
 *                                                                         *
 *  // setup theoretical model and experimental data:                      *
 *  chi_square.model ( model_ptr );                                        *
 *  chi_square.data ( x_ptr, y_ptr, dy_ptr, n );                           *
 *                                                                         *
 *  // choose MINUIT minimizer if necessary:                               *
 *  chi_square.strategy ( strategy );                                      *
 *                                                                         *
 *  // perform fitting:                                                    *
 *  chi_square.minimize ( robust, verbose );                               *
 *                                                                         *
 *  // produce some output:                                                *
 *  cout << chi_square;                                                    *
 *                                                                         *
 ***************************************************************************/

#include <QTextStream>

class ChiSquare : private TMinuit {

public: /** C-tor & d-tor */

  ChiSquare();
  ~ChiSquare();

public: /** Interface */

  /** Current model */
  const Model & model() const;
  /** Setup model */
  void model ( Model * m );
  /** Data access (write) */
  void data ( const double * x, const double * y, const double * dy, 
	      const unsigned int n );
  /** Chi-square value (to be minimized) */
  const double & value() const;
  /** Chi-square norm */
  const double norm() const;
  /** MINUIT interface */
  void minimize ( bool robust = false, const int verbose = -1 );
  /** Current MINUIT minimizer */
  const string & strategy() const;
  /** Setup MINUIT minimizer. Strategies are: 
      MINIMIZE (default), MIGRAD, SCAN, SEEK, SIMPLEX */
  bool strategy ( const string & method );

private: /** Functions to be minimized */

  /** Standard Chi-square value */
  const double & standard() const;
  /** Weighted Chi-square value */
  const double & weighted() const;

private: /** Storages */

  /** Model storage */
  mutable Model * __m_ptr;
  /** Data storages */
  const double * __x_ptr;
  const double * __y_ptr;
  const double * __dy_ptr;
  unsigned int __n;
  /** Pointer to function to be minimized */
  typedef const double & (ChiSquare::*fun4min) () const;
  fun4min __fun_ptr;
  /** Current Chi-square value */
  mutable double __chi2;

private: /** Robustness stuff */

  /** Weights for weighted LSF value */
  vector<double> __weight;
  /** Norm (sum of weights) for weighted LSF value */
  double __norm;
  /** Calculate Tukey's bi-square weights for weighted LSF value */
  void set_weights();
  /** Set weights = 1 for weighted LSF value 
      (should be called before any weights usage!) */
  void init_weights();
  /** Constant for Tukey's bi-weights calculation */
  double C_TUKEY;

private: /** MINUIT stuff (defined in lsf_minuit.cpp) */

  /** Static pointer to this for FCN */
  static const ChiSquare * __lsf_ptr;
  /** User FCN function for MINUIT */
  static void fcn ( int &, double *, double &, double *, int );
  /** Set printout level for MINUIT */
  void set_mn_printout ( const int level = -1 );
  /** Set parameters for MINUIT */
  void set_mn_parms();
  /** Get parameters from MINUIT */
  void get_mn_parms() const;
  /** Current MINUIT strategy */
  string __strategy;
  /** Available MINUIT strategies */
  static const string __strategies[];

};

/** ChiSquare output */
ostream & operator<< ( ostream & out, const ChiSquare & chi );
QTextStream & operator<< ( QTextStream & out, const ChiSquare & chi );

#endif // MINIMIZE_H
