/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Tue Mar  5 19:02:11 MSK 2002
    copyright            : (C) 2002 by Alexei Soloviev, Alexei Stadnik
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "mainwindow.h"

#include "stdmodel.h"
#include "sansmodel.h"
#include "yumomodel.h"

#include "lsf.h"

#include <vector>
#include <string>
using namespace std;

#include <QApplication>
#include <QStyleFactory>
#include <QCommandLineParser>
#include <QFile>

int main ( int argc, char * argv[] )
{
    if ( argc == 1 )
    {
        QApplication app(argc, argv);
        MainWindow w;
        w.show();

        QApplication::setStyle(QStyleFactory::create("fusion"));

        return app.exec();
    }

    QCoreApplication app(argc, argv);
    QCoreApplication::setApplicationName(PACKAGE);
    QCoreApplication::setApplicationVersion(VERSION);

    QCommandLineParser parser;

    parser.addHelpOption();
    parser.addVersionOption();

    parser.addOptions({
                          // Model
                          {"std", "Standard models", "model"},
                          {"sans", "SANS models", "model"},
                          {"yumo", "SANS models for YuMO spectrometer", "model"},
                          // Processing
                          {{"s", "strategy"}, "Change MINUIT strategy [MINIMIZE]", "value"},
                          {{"R", "range"}, "Fitting in range x1:x2 [none]", "value"},
                          {{"r", "robust"}, "Robust fitting [no]"},
                          {{"p", "printout"}, "Output MINUIT processing information [none]", "value"},
                          {"rc-read", "Read rc-file [no]"},
                          {"rc-write", "Write rc-file [no]"},
                          // Output
                          {{"o", "output"}, "Output data file [none]", "file"},
                          {{"H", "header"}, "Output results file [standard output]", "file"},
                      });

    parser.process(app);


    QString optarg;

    QTextStream cout(stdout);
    QTextStream cerr(stderr);

    /** Output stuff */
    QTextStream * info = &cout;
    QTextStream * data = 0;
    QTextStream * verb = &cerr;

    optarg = parser.value("output");
    if (!optarg.isEmpty())
    {
        data = new QTextStream ( new QFile ( optarg ) );
        data->device()->open(QIODevice::WriteOnly);
    }
    optarg = parser.value("header");
    if (!optarg.isEmpty())
    {
        info = new QTextStream ( new QFile ( optarg ) );
        info->device()->open(QIODevice::WriteOnly);
    }

    /** Data stuff */
    vector<double> x, y, dy, dx;
    double * x_min = 0;
    double * x_max = 0;
    optarg = parser.value("range");
    if (!optarg.isEmpty())
    {
        QStringList range = optarg.split(":");
        if ( range.count() != 2 ) {
            *verb << PACKAGE << ": wrong range argument" << endl;
            return 1;
        }
        if (!range.first().isEmpty())
            x_min = new double(range.first().toDouble());
        if (!range.last().isEmpty())
            x_max = new double(range.last().toDouble());
        if ( x_min && x_max && *x_min >= *x_max ) {
            *verb << PACKAGE << ": can't fit with an empty range!" << endl;
            return 1;
        }
    }

    /** Model stuff */
    Model * model = 0;
    optarg = parser.value("std");
    if (!optarg.isEmpty())
        model = model_ptr<STDModel> ( optarg.toStdString() );
    optarg = parser.value("sans");
    if (!optarg.isEmpty())
        model = model_ptr<SANSModel> ( optarg.toStdString() );
    optarg = parser.value("yumo");
    if (!optarg.isEmpty())
        model = model_ptr<YuMOModel> ( optarg.toStdString() );

    bool do_rc_read = parser.isSet("rc-read");
    bool do_rc_write = parser.isSet("rc-write");

    /** Fitting stuff */
    string strategy = "MINIMIZE";
    optarg = parser.value("strategy");
    if (!optarg.isEmpty())
        strategy = optarg.toStdString();
    bool robust = parser.isSet("robust");
    int printout = -1;
    optarg = parser.value("printout");
    if (!optarg.isEmpty())
        printout = optarg.toInt();

    /** Input stuff */
    QTextStream * in = 0;
    QStringList notOptions = parser.positionalArguments();
    if ( notOptions.count() ) {
        in = new QTextStream ( new QFile ( notOptions.first() ) );
        in->device()->open(QIODevice::ReadOnly);
        if ( !in->device()->isOpen() ) {
            *verb << PACKAGE << ": cannot open input file \""
                  << notOptions.first() << "\"" << endl;
            return 1;
        }
        if ( notOptions.count() > 1 ) {
            *verb << PACKAGE << ": wrong arguments: ";
            *verb << notOptions.join(" ");
            *verb << endl;
            return 1;
        }
    } else {
        *verb << PACKAGE << ": no input file specified"<< endl;
        return 1;
    }

    if ( model == 0 ) {
        *verb << PACKAGE << ": no model specified"<< endl;
        return 1;
    }

    while ( !in->atEnd() ) {
        QStringList line = in->readLine().split(" ", QString::SkipEmptyParts);
        if ( line.count() < 3 ) {
            *verb << PACKAGE << ": wrong input file format"<< endl;
            return 1;
        }
        bool ok;
        double ax = line[0].toDouble(&ok);
        double ay = line[1].toDouble(&ok);
        double ady = line[2].toDouble(&ok);
        if ( !ok ) continue;
        if ( x_min && ax < *x_min ) continue;
        if ( x_max && ax > *x_max ) continue;
        x.push_back(ax);
        y.push_back(ay);
        dy.push_back(ady);
        if ( line.count() > 3 ) {
            double adx = line[3].toDouble();
            dx.push_back(adx);
        }
    }
    if ( x.empty() ) {
        *verb << PACKAGE << ": all points undefined!"<< endl;
        return 1;
    }

    if ( YuMOModel *y = dynamic_cast<YuMOModel*>(model) ) {
        if ( !y->initResolution ( &*x.begin(), &*dx.begin(), dx.size() ) ) {
            *verb << PACKAGE << ": input file containes no resolution data"<< endl;
        }
    }

    ChiSquare chi_square;
    chi_square.model ( model );
    chi_square.data ( &*x.begin(), &*y.begin(), &*dy.begin(), x.size() );

    if ( !chi_square.strategy ( strategy ) ) {
        *verb << PACKAGE << ": wrong minimizer" << endl;
        return 1;
    }

    if ( do_rc_read ) model->rc_read();
    chi_square.minimize ( robust, printout );
    if ( do_rc_write ) model->rc_write();

    *info << chi_square;

    if ( data && data->device()->isOpen() ) {
        data->setRealNumberNotation(QTextStream::ScientificNotation);
        data->setFieldAlignment(QTextStream::AlignRight);
        data->setRealNumberPrecision(4);

        data->setFieldWidth(12);
        for ( vector<double>::const_iterator p = x.begin(); p != x.end(); ++p ) {
            *data << *p << (*model)(*p) << endl;
        }
    }

    return 0;
}
