/* -*- C++ -*- *************************************************************
 *
 *  model.h, Tue Mar 5 2002
 *  Copyright (C) 2002 Soloviev Alexei <solovjev@cv.jinr.ru>, Alexei Stadnik
 *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MODEL_H
#define MODEL_H

typedef int size_type;

/**Model base class
  *@author Alexei Soloviev, Alexei Stadnik
  */

#include "parameter.h"

#include <vector>
#include <string>
#include <cstdlib>
using namespace std;

class Model {
public: 
  Model ( const vector<Parameter> & rhs, const string & name );
  virtual ~Model ();

  /** Number of parameters */
  virtual size_type size() const;

  /** Access to parameter no. index (write) */
  virtual double & operator[] ( size_type index) ;
  /** Access to parameter no. index (read) */
  virtual const double & operator[] ( size_type index ) const;

  /** Access to error of parameter no. index (write) */
  virtual double & error ( size_type index) ;
  /** Access to error of parameter no. index (read) */
  virtual const double & error ( size_type index ) const;

  /** Access to min value of parameter no. index (read) */
  virtual const double & min ( size_type index ) const;
  /** Access to max value of parameter no. index (read) */
  virtual const double & max ( size_type index ) const;

  /** Name of parameter no. index */
  virtual const string & name ( size_type index ) const;

  /** Theoretical function */
  virtual double operator() ( const double& x ) const = 0;
  /** Model name */
  virtual const string & name() const;

  /** Read model parameters from rc-file */
  void rc_read();
  /** Write model parameters to rc-file */
  void rc_write() const;

protected: // Protected attributes
  /** Model parameters */
  vector<Parameter> fParm;
  /** Current model name */
  string fName;
};


/** Some useful templates.
  * @author Alexei Soloviev
  */

#include <map>
#include <iostream>
using namespace std;

/** Model dictionary.
  * Meaning: dictionary["name"] = constructor
  */
template < class WorkModel >
class Dictionary : public map < string, const WorkModel (*) () > {};

/** Model usage.
  * Note: static const Dictionary<WorkModel> & dictionary();
  * should be defined in an inherited class to use this template.
  */
template < class WorkModel >
void usage ( ostream & out = cerr )
{
  static Dictionary<WorkModel> fDictionary = WorkModel::dictionary();
  out << "Available models: " << endl;
  for ( typename Dictionary<WorkModel>::const_iterator p = fDictionary.begin();
	p != fDictionary.end(); ++p ) {
    out << '\t' << p->first << '\t' << (p->second)().name() << endl;
  }
  exit(1);
}

/** Model can be construct over comand line parameters.
  * Note: static const Dictionary<WorkModel> & dictionary();
  * should be defined in an inherited class to use this template.
  */
template < class WorkModel >
WorkModel * model_ptr ( const string & name )
{
  if ( name == "help" ) usage<WorkModel> ();
  static Dictionary<WorkModel> dict = WorkModel::dictionary();
  if ( dict.find( name ) == dict.end() ) return 0;
  static WorkModel model = dict[name]();  // only ONE model of this type
  return &model;
}

template < class WorkModel >
WorkModel & model_ref ( const string & name )
{
  return *model_ptr<WorkModel>(name);
}

#endif
