/***************************************************************************
                          sansmodel.cpp  -  description
                             -------------------
    begin                : Sat Mar 9 2002
    copyright            : (C) 2002 by Alexei Soloviev, Alexei Stadnik
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "sansmodel.h"
#include "bessel.h"

#include <cmath>

static const double SANS_MIN_SIZE = 0;
static const double SANS_MAX_SIZE = 10000;

SANSModel::SANSModel( const std::vector<Parameter> & p, 
		      const IF & f1, const TF & f2, 
		      const std::string & n )
  : Model(p,n), Integrand(), fIntegFuncPtr(f1), fTheorFuncPtr(f2)
{}

SANSModel::~SANSModel()
{}

/** Theoretical function */
double SANSModel::operator() ( const double & x)  const {
  return ( this->*fTheorFuncPtr ) ( x );
}

/** Function to be integrated */
double SANSModel::operator() ( const double * x ) const {
  return ( this->*fIntegFuncPtr ) ( x );
}

/** To construct model over comand line parameters */
const Dictionary<SANSModel> & SANSModel::dictionary() {
  static Dictionary<SANSModel> fDictionary;
  if ( fDictionary.empty() ) {
    fDictionary["b"] = &ball;
    fDictionary["e2"] = &ellipsoid2;
    fDictionary["e3"] = &ellipsoid3;
    fDictionary["c"] = &cylinder;
    fDictionary["c2"] = &cylinder2;
    fDictionary["p"] = &parallelepiped;
    fDictionary["ss"] = &sphericalShell;
    fDictionary["e2s"] = &ellipsoid2Shell;
    fDictionary["e3s"] = &ellipsoid3Shell;
  }
  return fDictionary;
}

/** Ball constructor */
const SANSModel SANSModel::ball() {
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Radius", 70, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm, 
		     &SANSModel::ball, &SANSModel::ball, 
		     "Ball" );
}
/** Ball function to be integrated and its theoretical function */
double SANSModel::ball( const double * ) const{
  double arg = __Q * fParm[0].value();
  double res = 1;
  if ( arg ) res *= 3 * ( sin(arg) - arg * cos(arg) ) / ( arg * arg * arg );
  return res *= res;
}
double SANSModel::ball( const double & x ) const{
  __Q = x;
  return operator() ( &x )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Ellipsoid-2 constructor */
const SANSModel SANSModel::ellipsoid2(){
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Axis", 30, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Excentricity", 1, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm, 
		     &SANSModel::ellipsoid2, &SANSModel::ellipsoid2, 
		     "Two axis ellipsoid" );
}
/** Ellipsoid-2 function to be integrated and its theoretical function */
double SANSModel::ellipsoid2(const double * x) const{
  double arg = __Q * fParm[0].value()
    * sqrt( 1 + x[0]*x[0] 
	    * ( fParm[1].value()*fParm[1].value() - 1 ) );
  double res = 1;
  if ( arg ) res = 3 * ( sin(arg) - arg * cos(arg) ) / ( arg * arg * arg );
  return res *= res;
}
double SANSModel::ellipsoid2(const double & x) const{
  __Q = x;
  return integral ( *this, 0, 1 )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Ellipsoid-3 constructor */
const SANSModel SANSModel::ellipsoid3(){
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "a", 70, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "b", 50, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "c", 30, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm, 
		     &SANSModel::ellipsoid3, &SANSModel::ellipsoid3, 
		     "Three axial ellipsoid" );
}
/** Ellipsoid-3 function to be integrated and its theoretical function */
double SANSModel::ellipsoid3(const double * x) const{
  double aCos = fParm[0].value() * cos ( M_PI / 2 * x[0] );
  double bSin = fParm[1].value() * sin ( M_PI / 2 * x[0] );
  double arg = __Q
    * sqrt( ( aCos*aCos + bSin*bSin ) * ( 1 - x[1]*x[1] )
	    + fParm[2].value()*fParm[2].value() * x[1]*x[1] ); 
  double res = 1;
  if ( arg ) res = 3 * ( sin(arg) - arg * cos(arg) ) / ( arg * arg * arg );
  return res *= res;
}
double SANSModel::ellipsoid3(const double & x) const{
  __Q = x;
  static const double a[] = {0,0}, b[] = {1,1};
  return integral ( *this, 2, a, b ) 
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Cylinder constructor */
const SANSModel SANSModel::cylinder(){
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Radius", 170, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Length", 100, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm, 
		     &SANSModel::cylinder, &SANSModel::cylinder, 
		     "Cylinder" );
}
/** Cylinder function to be integrated and its theoretical function */
double SANSModel::cylinder(const double * x) const{
  double arg, res = 1;

  arg = __Q * fParm[0].value() * sqrt( 1 - x[0]*x[0] );
  if ( arg ) res *= ( 2 * bessj1(arg) / arg );

  arg = __Q * fParm[1].value() * x[0] / 2;
  if ( arg ) res *= ( sin(arg) / arg );

  return res *= res;
}
double SANSModel::cylinder(const double & x) const{
  __Q = x;
  return integral ( *this, 0, 1 )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Cylinder-2 constructor */
const SANSModel SANSModel::cylinder2(){
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Axis", 100, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Excentricity", 1, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Length", 35, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm, 
		     &SANSModel::cylinder2, &SANSModel::cylinder2,
		     "Elliptical cylinder" );
}
/** Cylinder-2 function to be integrated and its theoretical function */
double SANSModel::cylinder2(const double * x) const{
  double v = fParm[1].value() * fParm[1].value();

  double arg, res = 1;

  arg = __Q * fParm[0].value()
    * sqrt( ( 1 - x[0]*x[0] ) * 
	    ( (1 + v) + (1 - v) * cos(x[1]*M_PI) ) / 2 );
  if ( arg ) res *= ( 2 * bessj1(arg) / arg );
  arg = __Q * fParm[2].value() * x[0] / 2;
  if ( arg ) res *= ( sin(arg) / arg );

  return res *= res;
}
double SANSModel::cylinder2(const double & x) const{
  __Q = x;
  static const double a[] = {0,0}, b[] = {1,1};
  return integral ( *this, 2, a, b ) 
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Parallelepiped constructor */
const SANSModel SANSModel::parallelepiped(){
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "A", 200, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "B", 1, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "C", 30, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm, 
		     &SANSModel::parallelepiped, &SANSModel::parallelepiped, 
		     "Parallelepiped" );
}
/** Parallelepiped function to be integrated and its theoretical function */
double SANSModel::parallelepiped(const double * x) const{
  double arg, res = 1;

  double y = x[1]*M_PI/2;
  arg = __Q * fParm[0].value() * sin(y) / 2;
  if ( arg ) res *= ( sin(arg) / arg );
  arg = __Q * fParm[1].value() * sqrt( 1 - x[0]*x[0] ) * cos(y) / 2;
  if ( arg ) res *= ( sin(arg) / arg );
  arg = __Q * fParm[1].value() * fParm[2].value() * x[0] / 2;
  if ( arg ) res *= ( sin(arg) / arg );

  return res *= res;
}
double SANSModel::parallelepiped(const double & x) const{
  __Q = x;
  static const double a[] = {0,0}, b[] = {1,1};
  return integral ( *this, 2, a, b ) 
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Spherical shell constructor */
const SANSModel SANSModel::sphericalShell() {
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Outer_radius", 65, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Inner_radius", 20, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm,
		     &SANSModel::sphericalShell, &SANSModel::sphericalShell,
		     "Spherical shell" );
}
/** Spherical shell function to be integrated and its theoretical function */
double SANSModel::sphericalShell( const double * ) const{
  double arg1 = __Q * fParm[0].value();
  double arg2 = __Q * fParm[1].value();
  double res1 = 1, res2 = 1;
  if ( arg1 )
    res1 *= 3 * ( sin(arg1) - arg1 * cos(arg1) ) / ( arg1 * arg1 * arg1 );
  if ( arg2 )
    res2 *= 3 * ( sin(arg2) - arg2 * cos(arg2) ) / ( arg2 * arg2 * arg2 );
  arg2 = fParm[1].value() / fParm[0].value();
  res2 *= ( arg2 * arg2 * arg2 );
  res1 -= res2;
  return res1 *= res1;
}
double SANSModel::sphericalShell( const double & x ) const{
  __Q = x;
  return operator() ( &x )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Elipsoid-2 shell constructor */
const SANSModel SANSModel::ellipsoid2Shell() {
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Rotation_semiaxis_of_core", 60, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Semiaxis_of_core", 60, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Thickness_of_shell", 30, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Scattering_length_density_of_core", 6e10, 0.1 ) );
  parm.push_back ( Parameter( "Scattering_length_density_of_shell", 2e10, 0.1 ) );
  parm.push_back ( Parameter( "Scattering_length_density_of_solvent", 6e10, 0.1 ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm,
		     &SANSModel::ellipsoid2Shell, &SANSModel::ellipsoid2Shell,
		     "Two axis ellipsiodal coreshell" );
}
/** Ellipsoid-2 shell function to be integrated and its theoretical function */
double SANSModel::ellipsoid2Shell( const double * x) const{
  double xc = __Q * sqrt ( fParm[0].value()*fParm[0].value() * x[0]*x[0] +
			   fParm[1].value()*fParm[1].value() * ( 1 - x[0]*x[0] ) );
  double xt = __Q * sqrt ( (fParm[0].value()+fParm[2].value())*(fParm[0].value()+fParm[2].value()) * x[0]*x[0] +
			   (fParm[1].value()+fParm[2].value())*(fParm[1].value()+fParm[2].value()) * ( 1 - x[0]*x[0]) );

  double Vc = 4.0 / 3.0 * M_PI * fParm[0].value() * fParm[1].value()*fParm[1].value();
  double Vt = 4.0 / 3.0 * M_PI * (fParm[0].value()+fParm[2].value()) *
    (fParm[1].value()+fParm[2].value())*(fParm[1].value()+fParm[2].value());

  double res1 = ( fParm[3].value() - fParm[4].value() ) * Vc;
  if(xc)
    res1 *= 3 * ( sin(xc) - xc * cos(xc) ) / ( xc * xc * xc );

  double res2 = ( fParm[4].value() - fParm[5].value() ) * Vt;
  if(xt)
    res2 *= 3 * ( sin(xt) - xt * cos(xt) ) / ( xt * xt * xt );

  res1 += res2;
  return res1 *= res1;
}
double SANSModel::ellipsoid2Shell( const double & x ) const{
  __Q = x;
  return integral ( *this, 0, 1 )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Elipsoid-3 shell constructor */
const SANSModel SANSModel::ellipsoid3Shell() {
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Semiaxis_a_of_core", 70, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Semiaxis_b_of_core", 50, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Semiaxis_c_of_core", 30, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Thickness_of_shell", 30, 0.1, SANS_MIN_SIZE, SANS_MAX_SIZE ) );
  parm.push_back ( Parameter( "Scattering_length_density_of_core", 6e10, 0.1 ) );
  parm.push_back ( Parameter( "Scattering_length_density_of_shell", 2e10, 0.1 ) );
  parm.push_back ( Parameter( "Scattering_length_density_of_solvent", 6e10, 0.1 ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0, 0.1 ) );
  return SANSModel ( parm,
		     &SANSModel::ellipsoid3Shell, &SANSModel::ellipsoid3Shell,
		     "Triaxial ellipsiodal coreshell" );
}
/** Ellipsoid-3 shell function to be integrated and its theoretical function */
double SANSModel::ellipsoid3Shell( const double * x) const{
  double aCos1 = fParm[0].value() * cos ( M_PI / 2 * x[0] );
  double bSin1 = fParm[1].value() * sin ( M_PI / 2 * x[0] );
  double arg1 = __Q
    * sqrt( ( aCos1*aCos1 + bSin1*bSin1 ) * ( 1 - x[1]*x[1] )
	    + fParm[2].value()*fParm[2].value() * x[1]*x[1] ); 
  double res1 = fParm[4].value() - fParm[5].value();
  if ( arg1 ) res1 = 3 * ( sin(arg1) - arg1 * cos(arg1) ) / ( arg1 * arg1 * arg1 );

  double aCos2 = (fParm[0].value()+fParm[3].value()) * cos ( M_PI / 2 * x[0] );
  double bSin2 = (fParm[1].value()+fParm[3].value()) * sin ( M_PI / 2 * x[0] );
  double arg2 = __Q
    * sqrt( ( aCos2*aCos2 + bSin2*bSin2 ) * ( 1 - x[1]*x[1] )
	    + (fParm[2].value()+fParm[3].value())*(fParm[2].value()+fParm[3].value()) * x[1]*x[1] ); 
  double res2 = fParm[5].value() - fParm[6].value();
  if ( arg2 ) res2 = 3 * ( sin(arg2) - arg2 * cos(arg2) ) / ( arg2 * arg2 * arg2 );

  res1 += res2;
  return res1 *= res1;
}
double SANSModel::ellipsoid3Shell( const double & x ) const{
  __Q = x;
  static const double a[] = {0,0}, b[] = {1,1};
  return integral ( *this, 2, a, b )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}
