/***************************************************************************
                          sansmodel.cpp  -  description
                             -------------------
    begin                : Mon Jan 13 2003
    copyright            : (C) 2003 by Alexei Soloviev
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "yumomodel.h"
#include "gintegral.h"

#include <cmath>
#include <functional>

YuMOModel::YuMOModel ( const SANSModel & rhs )
  : SANSModel(rhs)
{
  fName += " with resolution";
}

YuMOModel::~YuMOModel()
{}

/** Theoretical function */
double YuMOModel::operator() ( const double & x ) const
{
  if ( fResolution.find(x) == fResolution.end() )
    return SANSModel::operator() ( x );
  fMemory.Q = x;
  fMemory.dQ = sqrt ( 2 * fResolution[x] );
  double Qmin = fMemory.Q - 3 * fMemory.dQ;
  double Qmax = fMemory.Q + 3 * fMemory.dQ;
  static const double sqrt_2_pi = sqrt ( 2*M_PI );
  return integral( bind1st( mem_fun( &YuMOModel::sans_theor_fun ), this ),
		   Qmin, Qmax ) / ( sqrt_2_pi * fMemory.dQ );
}

/** Function to be integrated */
double YuMOModel::sans_theor_fun ( const double x ) const
{
  double res = SANSModel::operator() ( x );
  double arg = ( fMemory.Q - x ) / fMemory.dQ; arg *= arg; arg /= -2;
  return res *= exp ( arg );
}

/** To construct model over comand line parameters */
const Dictionary<YuMOModel> & YuMOModel::dictionary() 
{
  static Dictionary<YuMOModel> fDictionary;
  if ( fDictionary.empty() ) {
    fDictionary["b"] = &ball;
    fDictionary["e2"] = &ellipsoid2;
    fDictionary["e3"] = &ellipsoid3;
    fDictionary["c"] = &cylinder;
    fDictionary["c2"] = &cylinder2;
    fDictionary["p"] = &parallelepiped;
    fDictionary["ss"] = &sphericalShell;
    fDictionary["e2s"] = &ellipsoid2Shell;
    fDictionary["e3s"] = &ellipsoid3Shell;
  }
  return fDictionary;
}

/** Ball constructor */
const YuMOModel YuMOModel::ball()
{
  return SANSModel::ball();
}

/** Ellipsoid-2 constructor */
const YuMOModel YuMOModel::ellipsoid2()
{
  return SANSModel::ellipsoid2();
}

/** Ellipsoid-3 constructor */
const YuMOModel YuMOModel::ellipsoid3()
{
  return SANSModel::ellipsoid3();
}

/** Cylinder constructor */
const YuMOModel YuMOModel::cylinder()
{
  return SANSModel::cylinder();
}

/** Cylinder-2 constructor */
const YuMOModel YuMOModel::cylinder2()
{
  return SANSModel::cylinder2();
}

/** Parallelepiped constructor */
const YuMOModel YuMOModel::parallelepiped()
{
  return SANSModel::parallelepiped();
}

/** Spherical shell constructor */
const YuMOModel YuMOModel::sphericalShell()
{
  return SANSModel::sphericalShell();
}

/** Ellipsoid-2 shell constructor */
const YuMOModel YuMOModel::ellipsoid2Shell()
{
  return SANSModel::ellipsoid2Shell();
}

/** Ellipsoid-3 shell constructor */
const YuMOModel YuMOModel::ellipsoid3Shell()
{
  return SANSModel::ellipsoid3Shell();
}

/** Init Q-resolution dependency */
bool YuMOModel::initResolution ( const double * Q_ptr,
				 const double * dQ_ptr,
				 const size_type n ) const
{
  fResolution.clear();
  if ( !n ) return false;
  for ( size_type i = 0; i < n; ++i ) {
    fResolution[*Q_ptr++] = *dQ_ptr++;
  }
  return true;
}
