#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <QtGui>
#include <QFileDialog>
#include <QMessageBox>
#include <QWidgetAction>
#include <QToolButton>
#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
#  include <QPrintDialog>
#else
#  include <QtPrintSupport/QPrintDialog>
#endif

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "plot.h"

#include "stdmodel.h"
#include "sansmodel.h"
#include "yumomodel.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    process(this)
{
    ui->setupUi(this);

    process.setProcessChannelMode(QProcess::SeparateChannels);

    connect(&process, SIGNAL(readyReadStandardOutput()), this, SLOT(processReadyReadStandardOutput()));
    connect(&process, SIGNAL(readyReadStandardError()), this, SLOT(processReadyReadStandardError()));
    connect(&process, SIGNAL(finished(int)), this, SLOT(processFinished()));


    connect(ui->actionSave, SIGNAL(triggered()), this, SLOT(save()));
    connect(ui->actionPrint, SIGNAL(triggered()), this, SLOT(print()));

    connect(ui->actionAbout, SIGNAL(triggered()), this, SLOT(about()));
    connect(ui->actionSite, SIGNAL(triggered()), this, SLOT(site()));
    connect(ui->actionExit, SIGNAL(triggered()), this, SLOT(close()));

    runAction = new QAction("Run", ui->menuBar);
    connect(runAction, SIGNAL(triggered()), this, SLOT(run()));

    QList<QAction*> barActions = ui->menuBar->actions();
    barActions.insert(barActions.count()-1, runAction);
    ui->menuBar->clear();
    ui->menuBar->addActions(barActions);

    QActionGroup * modelGroup = new QActionGroup(ui->menuModel);
    modelGroup->setExclusive(true);
    {
        ui->menuSTD->clear();
        const Dictionary<STDModel> & dict = STDModel::dictionary();
        for ( Dictionary<STDModel>::const_iterator p = dict.begin(); p != dict.end(); ++p ) {
            QAction * a = ui->menuSTD->addAction((p->second)().name().c_str());
            a->setToolTip((p->first).c_str());
            a->setCheckable(true);
            connect(a, SIGNAL(toggled(bool)), this, SLOT(updateStatus()));
            modelGroup->addAction(a);
        }
    }
    {
        ui->menuSANS->clear();
        const Dictionary<SANSModel> & dict = SANSModel::dictionary();
        for ( Dictionary<SANSModel>::const_iterator p = dict.begin(); p != dict.end(); ++p ) {
            QAction * a = ui->menuSANS->addAction((p->second)().name().c_str());
            a->setToolTip((p->first).c_str());
            a->setCheckable(true);
            connect(a, SIGNAL(toggled(bool)), this, SLOT(updateStatus()));
            modelGroup->addAction(a);
        }
    }
    {
        ui->menuYuMO->clear();
        const Dictionary<YuMOModel> & dict = YuMOModel::dictionary();
        for ( Dictionary<YuMOModel>::const_iterator p = dict.begin(); p != dict.end(); ++p ) {
            QAction * a = ui->menuYuMO->addAction((p->second)().name().c_str());
            a->setToolTip((p->first).c_str());
            a->setCheckable(true);
            connect(a, SIGNAL(toggled(bool)), this, SLOT(updateStatus()));
            modelGroup->addAction(a);
        }
    }

    QActionGroup * strategyGroup = new QActionGroup(ui->menuStrategy);
    strategyGroup->setExclusive(true);
    foreach (QAction * a, ui->menuStrategy->actions())
    {
        connect(a, SIGNAL(toggled(bool)), this, SLOT(updateStatus()));
        strategyGroup->addAction(a);
    }

    QList<QString> rangeLabels = {"X min", "X max"};
    for (int i = 0; i < rangeLabels.count(); ++i)
    {
        QWidgetAction * rangeAcrion = new QWidgetAction(ui->menuRange);
        rangeAcrion->setText(rangeLabels[i]);
        rangeAcrion->setDefaultWidget(new QWidget(ui->menuRange));
        rangeAcrion->defaultWidget()->setLayout(new QHBoxLayout());
        rangeAcrion->defaultWidget()->layout()->addWidget(new QLabel(rangeLabels[i] + " = "));
        QLineEdit * lineEdit = new QLineEdit();
        lineEdit->setClearButtonEnabled(true);
        connect(lineEdit, SIGNAL(textChanged(QString)), this, SLOT(updateStatus()));
        rangeAcrion->defaultWidget()->layout()->addWidget(lineEdit);
        ui->menuRange->addAction(rangeAcrion);
    }

    QActionGroup * printoutGroup = new QActionGroup(ui->menuPrintout);
    printoutGroup->setExclusive(true);
    foreach (QAction * a, ui->menuPrintout->actions())
    {
        connect(a, SIGNAL(toggled(bool)), this, SLOT(updateStatus()));
        printoutGroup->addAction(a);
    }

    connect(ui->actionRobust, SIGNAL(triggered()), this, SLOT(updateStatus()));
    connect(ui->actionRC_read, SIGNAL(triggered()), this, SLOT(updateStatus()));
    connect(ui->actionRC_write, SIGNAL(triggered()), this, SLOT(updateStatus()));

    connect(ui->actionClear, SIGNAL(triggered()), this, SLOT(updateStatus()));

    connect(ui->actionOpen, SIGNAL(triggered()), this, SLOT(openFile()));

    fitFile = QFileInfo(QDir::temp(), ".fitter-data").absoluteFilePath();
    infoFile = QFileInfo(QDir::temp(), ".fitter-info").absoluteFilePath();


    ui->statusBar->addPermanentWidget(&status, 1);
    status.setReadOnly(true);
    status.setFrame(false);
    QPalette palette = status.palette();
    palette.setColor( QPalette::Base, Qt::transparent );
    status.setPalette(palette);

    executable = "fitter";

    reopenConfig();

    connect(ui->saveConfigPushButton, SIGNAL(clicked()), this, SLOT(saveConfig()));
    connect(ui->actionAllowEditConfig, SIGNAL(toggled(bool)), ui->configTab, SLOT(setEnabled(bool)));


    loadUIState();
    updateStatus();

    fitterUrl = QUrl("http://yumo.jinr.ru/software/fitter");

#ifdef Q_OS_WIN
    networkAccessManager = new QNetworkAccessManager(this);
    QObject::connect(networkAccessManager, SIGNAL(finished(QNetworkReply*)),
                     this, SLOT(replyFinished(QNetworkReply*)));

    checkForUpdate();
#endif
}

MainWindow::~MainWindow()
{
#ifdef Q_OS_WIN
    delete networkAccessManager;
#endif

    saveUIState();

    process.close();

    QFile::remove(fitFile);
    QFile::remove(infoFile);

    delete ui;
}

void MainWindow::saveUIState()
{
    QSettings settings(QSettings::IniFormat, QSettings::UserScope, "YuMO", QString(PACKAGE)+"_ui");
    settings.setIniCodec("UTF-8");

    settings.beginGroup(menuBar()->objectName());
    QList<QMenu *> l = menuBar()->findChildren<QMenu *>();
    foreach (QMenu * m, l)
    {
        settings.beginGroup(m->title());
        foreach (QAction * a, m->actions())
        {
            if (a->isEnabled() && a->isCheckable())
            {
                settings.setValue(a->text(), a->isChecked());
            }
            else if (dynamic_cast<QWidgetAction *>(a))
            {
                QWidget * w = dynamic_cast<QWidgetAction *>(a)->defaultWidget();
                QLineEdit * lineEdit = w->findChild<QLineEdit *>();
                if (lineEdit)
                {
                    settings.setValue(a->text(), lineEdit->text());
                }
            }
        }
        settings.endGroup();
    }
    settings.endGroup();

    settings.setValue(ui->tabWidget->objectName(), ui->tabWidget->currentIndex());

    foreach (Plot * plot, findChildren<Plot *>())
    {
        settings.beginGroup(plot->objectName());
        settings.setValue("xAxis", plot->xAxis->scaleType());
        settings.setValue("yAxis", plot->yAxis->scaleType());
        settings.endGroup();
    }

    settings.beginGroup("files");
    settings.setValue("currentFile",dataFile);
    settings.endGroup();

    settings.sync();
}

void MainWindow::loadUIState()
{
    QSettings settings(QSettings::IniFormat, QSettings::UserScope, "YuMO", QString(PACKAGE)+"_ui");
    settings.setIniCodec("UTF-8");

    settings.sync();

    settings.beginGroup(menuBar()->objectName());
    QList<QMenu *> l = menuBar()->findChildren<QMenu *>();
    foreach (QMenu * m, l)
    {
        settings.beginGroup(m->title());
        foreach (QAction * a, m->actions())
        {
            if (a->isEnabled() && a->isCheckable())
            {
                a->setChecked(settings.value(a->text(), a->isChecked()).toBool());
            }
            else if (dynamic_cast<QWidgetAction *>(a))
            {
                QWidget * w = dynamic_cast<QWidgetAction *>(a)->defaultWidget();
                QLineEdit * lineEdit = w->findChild<QLineEdit *>();
                if (lineEdit)
                {
                    lineEdit->setText(settings.value(a->text(), lineEdit->text()).toString());
                }
            }
        }
        settings.endGroup();
    }
    settings.endGroup();

    ui->tabWidget->setCurrentIndex(settings.value(ui->tabWidget->objectName(), ui->tabWidget->currentIndex()).toInt());

    foreach (Plot * plot, findChildren<Plot *>())
    {
        settings.beginGroup(plot->objectName());
        plot->xAxis->setScaleType((QCPAxis::ScaleType)settings.value("xAxis", plot->xAxis->scaleType()).toInt());
        plot->yAxis->setScaleType((QCPAxis::ScaleType)settings.value("yAxis", plot->yAxis->scaleType()).toInt());
        settings.endGroup();
    }

    settings.beginGroup("files");
    dataFile = settings.value("currentFile", QString()).toString();
    settings.endGroup();

    if (QFileInfo(dataFile).exists())
        drawData();
}

#ifdef Q_OS_WIN

void MainWindow::checkForUpdate()
{
    networkAccessManager->get(QNetworkRequest(fitterUrl));
}

static bool isUpdate(const QString & versionString)
{
    //return versionString != VERSION;

    QStringList digits = versionString.split(".");
    QStringList currentDigits = QString(VERSION).split(".");

    int n = std::max(digits.count(), currentDigits.count());
    while (digits.count() < n)
        digits.append("0");
    while (currentDigits.count() < n)
        currentDigits.append("0");

    for (int i = 0; i < n; ++i)
    {
        if (digits[i].toInt() != currentDigits[i].toInt())
            return (digits[i].toInt() > currentDigits[i].toInt());
    }

    return false;
}

void MainWindow::replyFinished(QNetworkReply * reply)
{
     // no error received?
    if (reply->error() == QNetworkReply::NoError)
    {
        QString package = QString(PACKAGE).toUpper();

        if (reply->url() == fitterUrl) // version
        {
            // read data from QNetworkReply here
            QByteArray bytes = reply->readAll();  // bytes
            QString text(bytes); // string

            QRegExp versionRegExp("<version>(.*)</version>");
            versionRegExp.indexIn(text);
            QString versionString = versionRegExp.cap(1);

            if (isUpdate(versionString))
            {
                int button = QMessageBox::question(this,
                                                   package + " update",
                                                   "Version " + versionString + " is available.\nWould you like to download it?");

                if (button == QMessageBox::Yes)
                {
                    QRegExp linkRegExp("<link>(.*)</link>");
                    linkRegExp.indexIn(text);
                    QString linkString = linkRegExp.cap(1);

                    networkAccessManager->get(QNetworkRequest(QUrl(linkString)));
                }
            }
        }
        else // download
        {
            const QString downloadsFolder = QStandardPaths::writableLocation(QStandardPaths::DownloadLocation);
            QFileInfo fileInfo = reply->url().path();
            QString path = downloadsFolder + QDir::separator() + fileInfo.fileName();
            QFile localFile(path);
            if (!localFile.open(QIODevice::WriteOnly))
                return;
            localFile.write(reply->readAll());
            localFile.close();

            int button = QMessageBox::question(this,
                                               package + " update",
                                               fileInfo.fileName() + " is downloaded.\nWould you like to install it?");

            if (button == QMessageBox::Yes)
            {
                if (QProcess::startDetached(path))
                    close();
                else
                    QMessageBox::critical(this,
                                          package + " update",
                                          "Failed to start " + fileInfo.fileName() + ".\nPlease try to run it manually.");
            }
        }
    }
    // Some http error received
    else
    {
        // handle errors here
    }

    // We receive ownership of the reply object
    // and therefore need to handle deletion.
    delete reply;
}

#endif

void MainWindow::updateStatus()
{
    command = executable;

    bool hasModel = false;
    foreach (QMenu * m, ui->menuModel->findChildren<QMenu *>(QString(), Qt::FindDirectChildrenOnly))
    {
        foreach (QAction * a, m->actions())
        {
            if (a->isCheckable() && a->isChecked())
            {
                command += m->toolTip() + "=" + a->toolTip();
                hasModel = true;
            }
        }
    }

    foreach (QAction * a, ui->menuStrategy->actions())
    {
        if (a->isCheckable() && a->isChecked())
        {
            command += " --strategy=" + a->toolTip();
        }
    }

    if (ui->actionRobust->isChecked())
        command += " --robust";

    QList<QLineEdit *> rangeLineEdits = ui->menuRange->findChildren<QLineEdit *>();
    QString xMin = rangeLineEdits.first()->text();
    QString xMax = rangeLineEdits.last()->text();
    if ( !xMin.isEmpty() || !xMax.isEmpty() ) command += " -R " + xMin + ":" + xMax;

    if (ui->actionRC_read->isChecked())
        command += " --rc-read";
    if (ui->actionRC_write->isChecked())
        command += " --rc-write";

    if (!dataFile.isEmpty())
        command += " \"" + dataFile + "\"";

    status.setText(command);

    command += " --output=\"" + fitFile + "\"";
    command += " --header=\"" + infoFile + "\"";

    int PRINTOUT = -1;
    foreach (QAction * a, ui->menuPrintout->actions())
    {
        if (a->isCheckable() && a->isChecked())
        {
            break;
        }
        ++PRINTOUT;
    }
    command += " --printout=" + QString::number(PRINTOUT);

    ui->statusBar->setToolTip(command);

    // run button
    runAction->setEnabled(!dataFile.isEmpty() && hasModel);
}

void MainWindow::drawData()
{
    ui->plot->clearGraphs();

    QString name = QFileInfo(dataFile).fileName();
    QString cols = "1:2:3";
    ui->plot->addGraphFromFile( dataFile, name, cols );
    //ui->plot->graph()->setLineStyle(QCPGraph::lsNone);

    ui->plot->rescaleAxes();
    ui->plot->replot();
}

void MainWindow::drawFit()
{
    QString name;
    foreach (QMenu * m, ui->menuModel->findChildren<QMenu *>(QString(), Qt::FindDirectChildrenOnly))
    {
        foreach (QAction * a, m->actions())
        {
            if (a->isCheckable() && a->isChecked())
            {
                name = a->text();
            }
        }
    }
    QString cols = "1:2";
    ui->plot->addGraphFromFile( fitFile, name, cols );

    ui->plot->rescaleAxes();
    ui->plot->replot();
}

void MainWindow::openFile()
{
    QSettings settings(QSettings::IniFormat, QSettings::UserScope, "YuMO", QString(PACKAGE)+"_ui");
    settings.setIniCodec("UTF-8");

    settings.sync();

    QString path = settings.value("files/InputDir", QDir::homePath()).toString();

    QFileDialog fileDialog(this,
                           tr("Open File"),
                           path,
                           tr("Data files (*.dat)"));
    fileDialog.setAcceptMode(QFileDialog::AcceptOpen);
    fileDialog.setFileMode(QFileDialog::ExistingFile);

    if (fileDialog.exec())
    {
        dataFile = fileDialog.selectedFiles().first();

        settings.setValue("files/InputDir", fileDialog.directory().absolutePath());
        settings.sync();
    }

    updateStatus();

    drawData();
    ui->infoBrowser->clear();
    ui->verbBrowser->clear();

    ui->actionSave->setEnabled(false);
    ui->actionPrint->setEnabled(false);
}

void MainWindow::run()
{
    if (process.state() == QProcess::Running || process.state() == QProcess::Starting)
    {
        process.close();
        runAction->setText("Run");
        return;
    }

    process.start(command);
    runAction->setText("Stop");

    if (ui->actionClear->isChecked())
    {
        drawData();
        ui->infoBrowser->clear();
    }
    ui->verbBrowser->clear();

    ui->actionSave->setEnabled(false);
    ui->actionPrint->setEnabled(false);
}

void MainWindow::processReadyReadStandardOutput()
{
    QString output = QString::fromLocal8Bit(process.readAllStandardOutput().constData());
    ui->verbBrowser->append(output);
}

void MainWindow::processReadyReadStandardError()
{
    QString title = QString(PACKAGE).toUpper() + " error";
    QString output = QString::fromLocal8Bit(process.readAllStandardError().constData());
    output = output.mid(output.indexOf(":")+1).trimmed();
    QMessageBox::critical(this, title, output);
}

void MainWindow::processFinished()
{
    runAction->setText("Run");

    if (process.exitStatus() == QProcess::CrashExit || process.exitCode() != 0 )
        return;

    drawFit();

    QFile file (infoFile);
    file.open(QFile::ReadOnly | QFile::Text);
    ui->infoBrowser->append(file.readAll());

    ui->actionSave->setEnabled(true);
    ui->actionPrint->setEnabled(true);

    reopenConfig();

    runAction->setEnabled(false); // prevent the same fit again
}

void MainWindow::save()
{
    QSettings settings(QSettings::IniFormat, QSettings::UserScope, "YuMO", QString(PACKAGE)+"_ui");
    settings.setIniCodec("UTF-8");

    settings.sync();

    QString path = settings.value("files/OutputDir", QDir::homePath()).toString();

    QFileDialog fileDialog(this,
                           tr("Save File"),
                           path,
                           tr("Fitter output files (*.fit)"));
    fileDialog.setAcceptMode(QFileDialog::AcceptSave);

    if (fileDialog.exec())
    {
        QStringList filesList = fileDialog.selectedFiles();
        QString fileName = filesList.last();

        settings.setValue("files/OutputDir", fileDialog.directory().absolutePath());
        settings.sync();

        QString extention = fileName.split(".",QString::SkipEmptyParts).last().toLower();
        if ( extention != "fit" )
            fileName += ".fit";

        if (QFile::exists(fileName))
            QFile::remove(fileName);
        QFile::copy(fitFile, fileName);

        QFileInfo fileInfo(fileName);
        QDir dir = fileInfo.absoluteDir();
        dir.mkdir("info");
        dir.cd("info");
        fileName = dir.absolutePath() + QDir::separator() + fileInfo.baseName() + ".info";

        if (QFile::exists(fileName))
            QFile::remove(fileName);
        QFile::copy(infoFile, fileName);

    }
}

void MainWindow::print()
{
    QPrinter printer;

    QPrintDialog *dialog = new QPrintDialog(&printer);
    dialog->setWindowTitle("Print Document");

    if (dialog->exec() != QDialog::Accepted)
        return;

    QPainter painter;
    painter.begin(&printer);

    QList<QWidget *> widgets = ui->tabWidget->currentWidget()->findChildren<QWidget *>();
    double xscale = printer.pageRect().width()/double(widgets.first()->width());
    double yscale = printer.pageRect().height()/double(widgets.first()->height());
    double scale = qMin(xscale, yscale);
    painter.scale(scale, scale);

    if (widgets.first() == ui->plot)
    {
        widgets.first()->render(&painter);
    }
    else
    {
        QTextEdit * textEdit = dynamic_cast<QTextEdit *>(widgets.first());
        if (textEdit)
            textEdit->document()->drawContents(&painter);
    }
}

void MainWindow::about()
{
    QString package = QString(PACKAGE).toUpper();
    QString version = QString(VERSION);

    QString windowTitle = "About " + package;
    QString title = "<b>" + package + " " + VERSION + "</b>";
    QString readme =
            "<p>Fitter is a C++ program aimed to fit a chosen theoretical\
            multi-parameter function through a set of data points.\
            The method of fitting is chi-square minimization.\
            Moreover, the robust fitting method can be applied in Fitter.\
            Fitter was disigned to be used for a small-angle neutron\
            scattering data analisis.\
            Respective theoretical models are implemented in it.\
            Some commonly used models (Gaussian and polynomials) are also\
            implemented for wider applicability.";

    QMessageBox::about(this, windowTitle, title + readme);
}

void MainWindow::site()
{
    QDesktopServices::openUrl(fitterUrl);
}

void MainWindow::saveConfig()
{
    QSettings settings(QSettings::IniFormat, QSettings::UserScope, "YuMO", PACKAGE);
    settings.setIniCodec("UTF-8");

    QFile file(settings.fileName());
    if (file.open(QFile::WriteOnly | QFile::Text))
    {
        QTextStream out(&file);
        out << ui->configTextEdit->document()->toPlainText();
        file.close();
    }

    settings.sync();

    updateStatus();
}

void MainWindow::reopenConfig()
{
    QSettings settings(QSettings::IniFormat, QSettings::UserScope, "YuMO", PACKAGE);
    settings.setIniCodec("UTF-8");

    settings.sync();

    QFile iniFile (settings.fileName());
    iniFile.open(QFile::ReadOnly | QFile::Text);

    QString text = QString::fromLocal8Bit(iniFile.readAll().constData());
    text.replace("%20", " ");

    ui->configTextEdit->setText(text);
    ui->configFilePathLabel->setText(iniFile.fileName());
}
