#include <QtGui>
#include <QMenu>
#include <QToolTip>


#ifdef HAVE_CONFIG_H
# include <config.h>
#else
# define PACKAGE "gluplot"
#endif

#include "plot.h"

const Plot::rgb
Plot::_easel[] = {
    { 0xff, 0x00, 0x00 }, /* color 01: red           */
    { 0x00, 0xc0, 0x00 }, /* color 02: dark green    */
    { 0x00, 0x80, 0xff }, /* color 03: dark blue     */
    { 0xc0, 0x00, 0xff }, /* color 04: dark magenta  */
    { 0x00, 0xee, 0xee }, /* color 05: cyan          */
    { 0xc0, 0x40, 0x00 }, /* color 06: orange        */
    { 0xee, 0xee, 0x00 }, /* color 07: yellow        */
    { 0x20, 0x20, 0xc0 }, /* color 08: royal blue    */
    { 0xff, 0xc0, 0x20 }, /* color 09: goldenrod1    */
};

const unsigned int
Plot::_easel_size = 9;

Plot::Plot(QWidget *parent) :
    QCustomPlot(parent),
    trimsZeros(true)
{
    // make range draggable and zoomable:
    setInteractions(QCP::iRangeDrag | QCP::iRangeZoom | QCP::iSelectLegend | QCP::iSelectPlottables | QCP::iSelectAxes | QCP::iMultiSelect);

    rubberBand = new QRubberBand(QRubberBand::Rectangle, this);

    // make top right axes clones of bottom left axes:
    axisRect()->setupFullAxesBox();
    // connect signals so top and right axes move in sync with bottom and left axes:
    connect(xAxis, SIGNAL(rangeChanged(QCPRange)), xAxis2, SLOT(setRange(QCPRange)));
    connect(yAxis, SIGNAL(rangeChanged(QCPRange)), yAxis2, SLOT(setRange(QCPRange)));
    connect(xAxis, SIGNAL(scaleTypeChanged(QCPAxis::ScaleType)), xAxis2, SLOT(setScaleType(QCPAxis::ScaleType)));
    connect(yAxis, SIGNAL(scaleTypeChanged(QCPAxis::ScaleType)), yAxis2, SLOT(setScaleType(QCPAxis::ScaleType)));
    connect(xAxis2, SIGNAL(scaleTypeChanged(QCPAxis::ScaleType)), xAxis, SLOT(setScaleType(QCPAxis::ScaleType)));
    connect(yAxis2, SIGNAL(scaleTypeChanged(QCPAxis::ScaleType)), yAxis, SLOT(setScaleType(QCPAxis::ScaleType)));

    xAxis->grid()->setSubGridVisible(true);
    yAxis->grid()->setSubGridVisible(true);

    xAxis->setNumberFormat("gb"); // e = exponential, b = beautiful decimal powers
    xAxis->setSubTickCount(10);

    yAxis->setNumberFormat("gb"); // e = exponential, b = beautiful decimal powers
    yAxis->setSubTickCount(10);

    // set locale to english, so we get english decimal separator:
    setLocale(QLocale(QLocale::English, QLocale::UnitedKingdom));

    // connect slot that ties some axis selections together (especially opposite axes):
    connect(this, SIGNAL(selectionChangedByUser()), this, SLOT(selectionChanged()));

    // setup policy and connect slot for context menu popup:
    setContextMenuPolicy(Qt::CustomContextMenu);
    connect(this, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(contextMenuRequest(QPoint)));

    connect(this, SIGNAL(mousePress(QMouseEvent*)), this, SLOT(onMousePress(QMouseEvent*)));
    connect(this, SIGNAL(mouseMove(QMouseEvent*)), this, SLOT(onMouseMove(QMouseEvent*)));
    connect(this, SIGNAL(mouseRelease(QMouseEvent*)), this, SLOT(onMouseRelease(QMouseEvent*)));
    connect(this, SIGNAL(mouseDoubleClick(QMouseEvent*)), this, SLOT(onMouseDoubleClick(QMouseEvent*)));
}

void Plot::addGraphFromFile ( const QString & filename,
                              const QString & name,
                              const QString & cols )
{
    QTextStream cerr (stderr);

    QFile in ( filename );
    in.open ( QFile::ReadOnly | QFile::Text );

    if ( !in.isOpen() ) {
        cerr << PACKAGE << ": can't read data file \""
             << filename << "\"" << endl;
        return;
    }

    // defining columns
    int xcol = 0, ycol = 0, dycol = 0, dxcol = 0;
    if ( cols.length() ) {
        QStringList columns = cols.split(":");
        unsigned int ncols = columns.count();
        switch ( ncols ) {
        case 1:
            ycol = columns[0].toInt();
            break;
        case 2:
            xcol = columns[0].toInt();
            ycol = columns[1].toInt();
            break;
        case 3:
            xcol = columns[0].toInt();
            ycol = columns[1].toInt();
            dycol = columns[2].toInt();
            break;
        case 4:
            xcol = columns[0].toInt();
            ycol = columns[1].toInt();
            dycol = columns[2].toInt();
            dxcol = columns[3].toInt();
            break;
        default:
            cerr << PACKAGE << ": " << "expecting "
                 << "'x:y', 'x:y:dy', 'x:y:dy:dx' or just 'y'"
                 << endl;
            return;
        }
    } else { // defaults
        xcol = 1, ycol = 2;
    }

    int l = 0;

    QVector<double> key, value, keyError, valueError;

    // reading columns to curve
    double x = 0, y = 0, dy, dx;
    while ( !in.atEnd() ) {
        QString line = in.readLine();
        if ( line.startsWith("#") ) continue;
        ++l;
        QStringList record = line.split(" ", QString::SkipEmptyParts);
        if ( xcol > record.count() ||
             ycol > record.count() ||
             dycol > record.count() ||
             dxcol > record.count() ) break;
        if ( xcol ) x = record[xcol-1].toDouble();
        else x += 1;
        if ( ycol ) y = record[ycol-1].toDouble();
        else y += 1;
        key.push_back(x);
        value.push_back(y);
        if ( dycol ) {
            dy = record[dycol-1].toDouble();
            valueError.push_back(dy);
        }
        if ( dxcol ) {
            dx = record[dxcol-1].toDouble();
            keyError.push_back(dx);
        }
    }
    in.close();

    if (!key.size())
        return;

    if (trimsZeros) {
        int indexBegin = 0;
        for ( ; indexBegin < value.size(); ++indexBegin ) {
            if ( value[indexBegin] != 0)
                break;
        }
        int indexEnd = value.size() - 1;
        for ( ; indexEnd >= 0; --indexEnd ) {
            if ( value[indexEnd] != 0)
                break;
        }
        int newSize = indexEnd - indexBegin + 1;
        key = key.mid(indexBegin, newSize);
        value = value.mid(indexBegin, newSize);
        keyError = keyError.mid(indexBegin, newSize);
        valueError = valueError.mid(indexBegin, newSize);
    }

    if (!key.size())
        return;

    addGraph();

    graph()->setName(name);

    graph()->setScatterStyle(QCPScatterStyle(QCPScatterStyle::ssDisc, 4));
    graph()->setErrorPen(QPen(QColor(180,180,180)));

    if (dycol && dxcol)
    {
        graph()->setDataBothError(key, value, keyError, valueError);
        graph()->setErrorType(QCPGraph::etBoth);
    }
    else if (dycol)
    {
        graph()->setDataValueError(key, value, valueError);
        graph()->setErrorType(QCPGraph::etValue);
    }
    else if (dxcol)
    {
        graph()->setDataKeyError(key, value, keyError);
        graph()->setErrorType(QCPGraph::etKey);
    }
    else
    {
        graph()->setData(key, value);
        graph()->setErrorType(QCPGraph::etNone);
        graph()->setScatterStyle(QCPScatterStyle(QCPScatterStyle::ssNone));
    }

    QColor graphColor (_easel[_color_index].r, _easel[_color_index].g, _easel[_color_index].b );
    graph()->setPen(graphColor);
    //graph()->setSelectedPen(graphColor);
    if (++_color_index >= _easel_size)
        _color_index = 0;

    legend->setVisible(true);
}

int Plot::clearGraphs()
{
    int res = QCustomPlot::clearGraphs();
    _color_index = 0;
    legend->setVisible(false);
    return res;
}

void Plot::contextMenuRequest(QPoint pos)
{
    QMenu *menu = new QMenu(this);
    menu->setAttribute(Qt::WA_DeleteOnClose);

    if (selectedGraphs().size() > 0) // general context menu on graphs requested
    {
        menu->addAction("Hide selected graph", this, SLOT(hideSelectedGraphs()));
        menu->addAction("Show selected graph", this, SLOT(showSelectedGraphs()));
    }
    else if (selectedAxes().count() > 0)
    {
        menu->addAction("Make selected axis Linear", this, SLOT(setSelectedAxisLinear()));
        menu->addAction("Make selected axis Logarithmic", this, SLOT(setSelectedAxisLogarithmic()));
    }
    else if (legend->selectTest(pos, false) >= 0) // context menu on legend requested
    {
        menu->addAction("Move to top left", this, SLOT(moveLegend()))->setData((int)(Qt::AlignTop|Qt::AlignLeft));
        menu->addAction("Move to top center", this, SLOT(moveLegend()))->setData((int)(Qt::AlignTop|Qt::AlignHCenter));
        menu->addAction("Move to top right", this, SLOT(moveLegend()))->setData((int)(Qt::AlignTop|Qt::AlignRight));
        menu->addAction("Move to bottom right", this, SLOT(moveLegend()))->setData((int)(Qt::AlignBottom|Qt::AlignRight));
        menu->addAction("Move to bottom left", this, SLOT(moveLegend()))->setData((int)(Qt::AlignBottom|Qt::AlignLeft));
    }

    menu->popup(mapToGlobal(pos));
}

void Plot::onMousePress(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton)
    {
        double x = xAxis->pixelToCoord(event->pos().x());
        double y = yAxis->pixelToCoord(event->pos().y());
        QToolTip::showText( mapToGlobal( event->pos() ), QString("%1, %2").arg(x).arg(y) );
    }

    if (event->button() == Qt::MiddleButton)
    {
        rubberOrigin = event->pos();
        rubberBand->setGeometry(QRect(rubberOrigin, QSize()));
        rubberBand->show();
    }
}

void Plot::onMouseMove(QMouseEvent *event)
{
    QToolTip::hideText();

    rubberBand->setGeometry(QRect(rubberOrigin, event->pos()).normalized());
}

void Plot::onMouseRelease(QMouseEvent *event)
{
    if (rubberBand->isVisible())
    {
        auto zoomRect = rubberBand->geometry();
        int xp1, yp1, xp2, yp2;
        zoomRect.getCoords(&xp1, &yp1, &xp2, &yp2);
        xAxis->setRange(xAxis->pixelToCoord(xp1), xAxis->pixelToCoord(xp2));
        yAxis->setRange(yAxis->pixelToCoord(yp1), yAxis->pixelToCoord(yp2));
        rubberBand->hide();
        replot();
    }
}

void Plot::onMouseDoubleClick(QMouseEvent *event)
{
    rescaleAxes();
}

void Plot::moveLegend()
{
    if (QAction* contextAction = qobject_cast<QAction*>(sender())) // make sure this slot is really called by a context menu action, so it carries the data we need
    {
        bool ok;
        int dataInt = contextAction->data().toInt(&ok);
        if (ok)
        {
            axisRect()->insetLayout()->setInsetAlignment(0, (Qt::Alignment)dataInt);
            replot();
        }
    }
}

void Plot::hideSelectedGraphs()
{
    for (int i=0; i<graphCount(); ++i)
    {
        if (graph(i)->selected())
        {
            graph(i)->setVisible(false);
            QColor textColor = legend->item(i)->textColor();
            textColor.setAlphaF(0.5);
            legend->item(i)->setTextColor(textColor);
        }
    }
    replot();
}

void Plot::showSelectedGraphs()
{
    for (int i=0; i<graphCount(); ++i)
    {
        if (graph(i)->selected())
        {
            graph(i)->setVisible(true);
            QColor textColor = legend->item(i)->textColor();
            textColor.setAlphaF(1);
            legend->item(i)->setTextColor(textColor);
        }
    }
    replot();
}

void Plot::selectionChanged()
{
    // synchronize selection of graphs with selection of corresponding legend items:
    for (int i=0; i<graphCount(); ++i)
    {
        QCPPlottableLegendItem *item = legend->itemWithPlottable(graph(i));
        if (item->selected() || graph(i)->selected())
        {
            item->setSelected(true);
            graph(i)->setSelected(true);
        }
    }
}

void Plot::setSelectedAxisLinear()
{
    foreach (QCPAxis * axis, selectedAxes())
    {
        axis->setScaleType(QCPAxis::stLinear);
    }
    rescaleAxes();
    replot();
}

void Plot::setSelectedAxisLogarithmic()
{
    foreach (QCPAxis * axis, selectedAxes())
    {
        axis->setScaleType(QCPAxis::stLogarithmic);
    }
    rescaleAxes();
    replot();
}
